/*
** fsbench.c - File System Benchmarking tool
**
** Copyright (c) 1999 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <ftw.h>
#include <time.h>
#include <math.h>

#include "petopt.h"

extern char version[];

int debug = 0;
int verbose = 0;

int bytes_per_file = 32*1024;
int files_per_dir = 8;
int dirs_per_dir  = 8;
int levels = 3;
int sync_files = 0;

uint64_t n_bytes = 0;
uint32_t n_files = 0;
uint32_t n_dirs  = 0;


void
program_header(FILE *fp)
{
    fprintf(fp, "[Fsbench, version %s (%s %s)]\n",
	    version, __DATE__, __TIME__);
}


void
usage(PETOPT *pp,
      FILE *fp)
{
    fputs("Usage: fsbench [OPTIONS] DIR\n", fp);
    fputs("\n", fp);
    petopt_usage(pp, fp);
}


int
parse_option(PETOPT *pp,
	     int opt,
	     char *arg)
{
    char *cp;
    int mult = 1;
    
    
    switch (opt)
    {
      case 'h':
	usage(pp, stdout);
	exit(EXIT_SUCCESS);

      case 'V':
	program_header(stdout);
	exit(EXIT_SUCCESS);

      case 's':
	sync_files = 1;
	break;
	
      case 'd':
	if (arg)
	    debug = atoi(arg);
	else
	    debug++;
	break;
	
      case 'v':
	if (arg)
	    verbose = atoi(arg);
	else
	    verbose++;
	break;

      case 'D':
	dirs_per_dir = atoi(arg);
	break;

      case 'F':
	files_per_dir = atoi(arg);
	break;

      case 'L':
	levels = atoi(arg);
	break;

      case 'B':
	cp = arg+strlen(arg)-1;
	if (cp > arg && isalpha(*cp))
	    switch (toupper(*cp))
	    {
	      case 'K':
		mult = 1024;
		break;

	      case 'M':
		mult = 1024*1024;
		break;

	      case 'G':
		mult = 1024*1024*1024;
		break;

	      default:
		return POE_INVALID;
	    }

	bytes_per_file = atoi(arg)*mult;
	break;
    }
    return 0;
}


int
create_file(const char *buf,
	    int size)
{
    int fd, len, i, saved_errno;
    char fdata[8192];
    
    
    fd = open(buf, O_CREAT|O_WRONLY, 0600);
    if (fd < 0)
	return errno;

    len = 0;
    while (len < size)
    {
	i = write(fd, fdata, size > sizeof(fdata) ? sizeof(fdata) : size);
	if (i < 0)
	{
	    saved_errno = errno;
	    close(fd);
	    return saved_errno;
	}

	len += i;
    }
	
    if (sync_files && fsync(fd) < 0)	
    {
	saved_errno = errno;
	close(fd);
	return saved_errno;
    }

	
    if (close(fd) < 0)
    {
	return errno;
    }

    return 0;
}


void
create_tree(const char *path,
	    int level)
{
    int i, fd, err;
    char buf[1024], *bufp;
    

    sprintf(buf, "%s/tempfile.", path);
    bufp = strlen(buf)+buf;
    
    for (i = 0; i < files_per_dir; i++)
    {
	sprintf(bufp, "%d", i);

	err = create_file(buf, bytes_per_file);
	if (err)
	{
	    fprintf(stderr, "fsbench: create_file(\"%s\", %d): %s\n",
		    buf, bytes_per_file, strerror(err));
	    exit(1);
	}
	n_files++;
	n_bytes += bytes_per_file;
    }
    
    sprintf(buf, "%s/tempdir.", path);
    bufp = strlen(buf)+buf;
    
    for (i = 0; i < dirs_per_dir; i++)
    {
	sprintf(bufp, "%d", i);
	
	if (mkdir(buf, 0700) < 0)
	{
	    fprintf(stderr, "fsbench: mkdir(\"%s\"): %s\n",
		    buf, strerror(errno));
	    exit(1);
	}

	n_dirs++;
	
	if (level < levels)
	    create_tree(buf, level+1);
    }
}


static int
destroy_object(const char *path,
	       const struct stat *sp,
	       int t,
	       struct FTW *ftwp)
{
    int err;

    
    switch (t)
    {
      case FTW_F:
      case FTW_SL:
      case FTW_SLN:
	if (unlink(path) < 0)
	{
	    fprintf(stderr, "fsbench: unlink(\"%s\"): %s\n",
		    path, strerror(errno), errno);
	    exit(1);
	}
	break;

      case FTW_D:
      case FTW_DP:
	if (rmdir(path) < 0)
	{
	    fprintf(stderr, "fsbench: rmdir(\"%s\"): %s\n",
		    path, strerror(errno));
	    exit(1);
	}
	break;

      case FTW_DNR:
	fprintf(stderr, "fsbench: %s: Unreadable\n", path);
	exit(1);

      case FTW_NS:
	fprintf(stderr, "fsbench: %s: Stat failed\n", path);
	exit(1);

      default:
	fprintf(stderr, "Unknown type: %d\n", t);
	exit(1);
    }

    return 0;
}


void
destroy_tree(const char *path)
{
    nftw(path, destroy_object, levels+10, FTW_DEPTH);
}

struct timespec
timespec_diff(const struct timespec *ts1,
	      const struct timespec *ts2)
{
    struct timespec res;
    int32_t nsec;

    
    nsec = ts1->tv_nsec - ts2->tv_nsec;

    res.tv_sec = ts1->tv_sec - ts2->tv_sec;
    if (nsec >= 0)
	res.tv_nsec = nsec;
    else
    {
	res.tv_sec -= 1;
	res.tv_nsec = ts2->tv_nsec - ts1->tv_nsec;
    }

    return res;
}

double
timespec_to_double(const struct timespec *ts)
{
    return (double) ts->tv_sec + (double) ts->tv_nsec / 1000000000.0;
}



static PETOPTS pov[] =
{
    { 'V', 0, 			"Version",	"Print version" },
    { 'h', 0,                   "HeLP",		"Print help" },
    { 'v', POF_OPT|POF_INT,	"Verbose",	"Enable verbose mode" },
    { 'd', POF_OPT|POF_INT, 	"DeBuG",	"Enable debug mode" },
    { 'D', POF_INT,		"Directories",  "Dirs per Dir to create" },
    { 'F', POF_INT,		"Files",	"Files per Dir to create" },
    { 'L', POF_INT,		"Levels",	"Subdir levels to create" },
    { 'B', POF_INT,		"Bytes",	"Bytes per File to create" },
    { 's', 0,			"Sync",		"Create files in Sync mode" },
    { -1, 0, NULL, NULL }
};

static PETOPT opts =
{
    0,
    &parse_option,
    NULL,
    pov
};



int
main(int argc,
     char *argv[])
{
    struct timespec create_start, create_stop;
    struct timespec destroy_start, destroy_stop;
    struct timespec d;
    int err, i;

    
    err = petopt_parse(&opts, argc, argv, &i);
    if (err)
	exit(1);

    if (i == argc)
    {
	fprintf(stderr, "fsbench: Missing DIR argument\n");
	exit(1);
    }

    if (verbose)
	program_header(stdout);

    if (verbose)
    {
	fputs("Creating tree...", stdout);
	fflush(stdout);
    }
    clock_gettime(CLOCK_REALTIME, &create_start);
    if (mkdir(argv[i], 0700) < 0)
    {
	fprintf(stderr, "fsbench: mkdir(\"%s\"): %s\n",
		argv[i], strerror(errno));
	exit(1);
    }
    create_tree(argv[i], 1);
    clock_gettime(CLOCK_REALTIME, &create_stop);
    
    if (verbose)
    {
	puts(" Done.");
	fputs("Destroying tree...", stdout);
    }
    
    clock_gettime(CLOCK_REALTIME, &destroy_start);
    destroy_tree(argv[i]);
    clock_gettime(CLOCK_REALTIME, &destroy_stop);
    
    if (verbose)
    {
	puts(" Done.");
	puts("\nResults:");
    }

    printf("Directories created/destroyed: %u\n", n_dirs);
    printf("Files created/destroyed:       %u\n", n_files);
    
    printf("Data created/destroyed:        ");
    if (n_bytes < 10*1024)
	printf("%lu bytes\n", (uint32_t) n_bytes);
    else
    {
	n_bytes /= 1024;
	if (n_bytes < 10*1024)
	    printf("%lu Kbytes\n", (uint32_t) n_bytes);
	else
	{
	    n_bytes /= 1024;
	    if (n_bytes < 10*1024)
		printf("%lu Mbytes\n", (uint32_t) n_bytes);
	    else
	    {
		n_bytes /= 1024;
		printf("%lu Gbytes\n", (uint32_t) n_bytes);
	    }
	}
    }

    d = timespec_diff(&create_stop, &create_start);
    printf("Creation time:            %10.2f\n", timespec_to_double(&d));

    d = timespec_diff(&destroy_stop, &destroy_start);
    printf("Destruction time:         %10.2f\n", timespec_to_double(&d));

    exit(0);
}
