/*
** petopt.c - Command line argument parser
**
** Copyright (c) 1999 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "petopt.h"




static int
compare_long(const char *user_option,
	     const char *aov_long,
	     int len)
{
    unsigned char *uc, *ac;
    int diff = 0;
    int saved_len = len;
    
    
    uc = (unsigned char *) user_option;
    ac = (unsigned char *) aov_long;
    
    while (len > 0 && diff == 0)
    {
	/* If we reach a word delimiter, skip till next word */
	if (*uc == '-')
	    while (*ac && *ac != '-')
		++ac;

	diff = (int) tolower(*uc) - (int) tolower(*ac);

	++uc;
	++ac;
	--len;
    }

    if (diff)
    {
	/* Check abbrev */
	int adiff = 0;
	
	
	uc = (unsigned char *) user_option;
	ac = (unsigned char *) aov_long;
	len = saved_len;

	while (len > 0 && adiff == 0)
	{
	    /* Locate abbrev characters (uppercase) */
	    while (*ac && (islower(*ac) || *ac == '-'))
		++ac;
	    
	    adiff = (int) tolower(*uc) - (int) tolower(*ac);

	    ++uc;
	    ++ac;
	    --len;
	}

	if (adiff == 0)
	    diff = 0;
    }
    return diff;
}



int
petopt_parse_option(struct petopt_parseinfo *pip,
		    int *optf,
		    char **optarg)
{
    int i;

    
    *optf = -1;
    *optarg = NULL;
    
    /* Index out of bounds? */
    if (pip->ai >= pip->argc)
    {
	return POE_EOF;
    }

    if (pip->ci == 0)
    {
	/* Non option or "-" */
	if (pip->argv[pip->ai][0] != '-' ||
	    (pip->argv[pip->ai][0] == '-' &&
	     pip->argv[pip->ai][1] == '\0'))
	{
	    return POE_EOF;
	}

	/* "--", stop parsing and skip */
	if (strcmp(pip->argv[pip->ai], "--") == 0)
	{
	    pip->ai++;
	    return POE_EOF;
	}

	if (pip->argv[pip->ai][1] == '-')
	{
	    /* Long option */
	    char *opt, *arg;
	    int len, ix;
	    
	    opt = pip->argv[pip->ai]+2;
	    arg = strchr(opt, '=');
	    if (arg)
		len = arg - opt;
	    else
		len = strlen(opt);

	    ix = -1;
	    for (i = 0; pip->pov[i].s != -1; i++)
	    {
		if (pip->pov[i].l != NULL &&
		    compare_long(opt, pip->pov[i].l, len) == 0)
		{
		    /* Match found */
		    if (ix < 0)
			ix = i;
		    else
		    {
			return POE_MULTI;
		    }
		}
	    }

	    
	    if (ix >= 0)
	    {
		pip->ai++;
		
		if (pip->pov[ix].f)
		{
		    /* Check for (optional) argument */
		    if (arg)
		    {
			/* Argument after "=" */
			*optarg = arg+1;
		    }
		    else if (pip->ai < pip->argc &&
			     (pip->argv[pip->ai][0] != '-' ||
			      pip->argv[pip->ai][0] == '\0'))
		    {
			/* Have an argument, empty string or "-" */
			if ((pip->pov[ix].f & POF_STR) ||
			    isdigit((int) pip->argv[pip->ai][0]))
			{
			    *optarg = pip->argv[pip->ai];
			    pip->ai++;
			}
		    }
		}

		/* Argument missing, and not optional? */
		if (pip->pov[ix].f && !(pip->pov[ix].f & POF_OPT) &&
		    *optarg == NULL)
		{
		    pip->ai--;
		    return POE_MISSING;
		}
		
		*optf = pip->pov[ix].s;
		return 0;
	    }

	    /* Unknown long command line switch */
	    return POE_OPTION;
	}
	else
	{
	    /* Short option */
	    
	    pip->ci = 1;
	}
    }

    /* Short option */

    i = 0;
    while (pip->pov[i].s != -1 && pip->pov[i].s != pip->argv[pip->ai][pip->ci])
	i++;

    if (pip->pov[i].s == -1)
    {
	/* Unknown short command line switch */
	
	*optf = pip->argv[pip->ai][pip->ci];
	return POE_OPTION;
    }
    
    pip->ci++;
    if (pip->argv[pip->ai][pip->ci] == '\0')
    {
	pip->ai++;
	pip->ci = 0;
    }

    /* Have an (optional) argument? */
    if (pip->pov[i].f)
    {
	if (pip->ai < pip->argc &&
	    (pip->ci != 0 ||
	     (pip->argv[pip->ai][0] != '-' ||
	      pip->argv[pip->ai][1] == '\0')))
	{
	    if ((pip->pov[i].f & POF_STR) ||
		isdigit((int) pip->argv[pip->ai][pip->ci]))
	    {
		*optarg = pip->argv[pip->ai]+pip->ci;
	    
		pip->ai++;
		pip->ci = 0;
	    }
	}
    }

    *optf = pip->pov[i].s;
    
    /* Argument missing, and not optional? */
    if (pip->pov[i].f && !(pip->pov[i].f & POF_OPT) &&
	*optarg == NULL)
    {
	return POE_MISSING;
    }
		
    return 0;
}


static int
petopt_print_error(PETOPT *pp,
		   int err,
		   char **argv,
		   int i,
		   int opt)
{
    switch (err)
    {
      case POE_EOF:
	return 0;

      case POE_OPTION:
	if (opt > 0)
	    fprintf(stderr, "%s: Unrecognized option: -%c\n",
		    argv[0],
		    opt);
	else
	    fprintf(stderr, "%s: Unrecognized option: %s\n",
		    argv[0],
		    argv[i]);
	break;

      case POE_MULTI:
	fprintf(stderr, "%s: Ambiguous option: %s\n",
		argv[0], argv[i]);
	break;
	
      case POE_MISSING:
	if (opt > 0)
	    fprintf(stderr, "%s: Missing argument for option: -%c\n",
		    argv[0], opt);
	else
	    fprintf(stderr, "%s: Missing argument for option: %s\n",
		    argv[0], argv[i]);
	break;

    	
      case POE_INVALID:
	if (opt > 0)
	    fprintf(stderr, "%s: Invalid argument for option: -%c\n",
		    argv[0], opt);
	else
	    fprintf(stderr, "%s: Invalid argument for option: %s\n",
		    argv[0], argv[i]);
	break;
    }

    return -1;
}


int
petopt_parse(PETOPT *pp,
	     int argc,
	     char *argv[],
	     int *oi)
{
    int err, opt;
    char *arg;
    struct petopt_parseinfo pib;

    
    pib.ai = 1;
    pib.ci = 0;
    pib.argc = argc;
    pib.argv = argv;
    pib.pov = pp->pov;

    
    while ((err = petopt_parse_option(&pib, &opt, &arg)) == 0)
	if ((err = pp->parse(pp, opt, arg)) != 0)
	{
	    *oi = pib.ai;
	    return err;
	}

    if (err == POE_EOF)
	err = 0;
    
    if (err)
    {
	if (pp->errprint)
	    err = pp->errprint(pp, err, pib.argv, pib.ai, opt);
	else
	    err = petopt_print_error(pp, err, pib.argv, pib.ai, opt);
    }
	
    *oi = pib.ai;
    return err;
}



int
petopt_usage(PETOPT *pp,
	     FILE *fp)
{
    int i, len;
    struct petopt_option *pov;


    pov = pp->pov;
    for (i = 0; pov[i].s != -1; i++)
    {
	if (pov[i].s > ' ' && pov[i].s <='~')
	    fprintf(fp, "  -%c, ", pov[i].s);
	else
	    fprintf(fp, "      ");

	len = 0;
	if (pov[i].l)
	{
	    len = fprintf(fp, "--%s", pov[i].l);
	    if (pov[i].f)
	    {
		if (pov[i].f & POF_OPT)
		    len += fprintf(fp, " [ARG]");
		else
		    len += fprintf(fp, " ARG");
	    }
	}
	
	while (len++ < 30)
	    putchar(' ');
	    
	if (pov[i].h)
	    fputs(pov[i].h, fp);
	putc('\n', fp);
    }

    return 0;
}
